
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.26.10',
    'branch_nick': 'release-1.26-LTS',
    'revision_id': '6fd89702aa872927639187208d7d6e6e0b352e7b',
    'revno': 10924
}
